#include "GameMath.hlsl"
#include "Shadows.hlsl"
#include "LightingFunctions.hlsl"
#include "GBuffer.hlsl"
#include "VoxelGI.hlsl"
#include "Shadows.hlsl"
//--------------------------------------------------------------------------------------
// Variables
//--------------------------------------------------------------------------------------
Texture2D txGB0 	: register(t0);
Texture2D txGB1 	: register(t1);
Texture2D txGB2 	: register(t2);
Texture2D txPrevFrame : register(t3);

#ifndef SAMLIN
#define SAMLIN
SamplerState samLinear : register(s0);
#endif
struct PSInput_Quad
{
    float4 pos : SV_Position;
    float4 texCoordOut : TEXCOORD0;
};

float3 GetUV(float3 position)
{
	float4 pV = mul(float4(position, 1.0f), View);
	float4 pVP=mul(pV,Projection);
	pVP.xy = float2(0.5f, 0.5f) + float2(0.5f, -0.5f) * pVP.xy/ pVP.w;
    return float3(pVP.xy, pV.z / pV.w);
}

float3 GetCameraSpacePos(float3 position)
{
    float4 pV = mul(float4(position, 1.0f), View);
    return pV.xyz / pV.w;
}
float distanceSquared(float2 a, float2 b)
{
    a -= b;
    return dot(a, a);
}

bool ScreenSpaceRT(float3 startPoint, float3 rayDir, float step, out float2 uv)
{
    float3 currentPos = startPoint + rayDir * step * 0.01;
    float3 clipSpacePos;
    float3 increment = rayDir * step;
    float lastDist = 100000;
    for (int i = 0; i < 32; i++)
    {
        currentPos += increment;
        clipSpacePos = GetUV(currentPos);
        // bound check
        if (clipSpacePos.x > 1 || clipSpacePos.y > 1 || /*clipSpacePos.z >= 1 || */
            clipSpacePos.x < 0 || clipSpacePos.y < 0 /*|| clipSpacePos.z <= 0*/)
        {
            uv = float2(0, 0);
            return false;
        }
        // to check if ray hit something we recalculate position(probably bad way to do this)
        float4 NormalSpec = txGB1.Sample(samLinear, clipSpacePos.xy);
        float ViewZ = DecodeFloatRG(NormalSpec.zw);
        ViewZ = ViewZ <= 0 ? fFarClip : ViewZ;
        float3 restoredPosition = GetUV(posFromDepth(ViewZ, clipSpacePos.xy).xyz);
        float dist = distance(restoredPosition.z, clipSpacePos.z);
        if (dist <= 0.1)
        {
            uv = clipSpacePos.xy;
            return true;
        }
        else
        {
            increment *= 1.05f;
        }
    }
    return false;
}

bool intersectsDepthBuffer(float z, float minZ, float maxZ)
{
    /*
     * Based on how far away from the camera the depth is,
     * adding a bit of extra thickness can help improve some
     * artifacts. Driving this value up too high can cause
     * artifacts of its own.
     */
    float depthScale = min(1.0f, z * 40);
    z += 0.5 + lerp(0.0f, 2.0f, depthScale);
    return (maxZ >= z) && (minZ - 0.5 <= z);
}
float linearDepthTexelFetch(float2 hitPixel)
{
    float4 NormalSpec = txGB1.Sample(samLinear, float2(hitPixel.x / fScreenWidth, 1 - hitPixel.y / fScreenHeight));
    return DecodeFloatRG(NormalSpec.zw);
}
void swap(inout float a, inout float b)
{
    float t = a;
    a = b;
    b = t;
}
// Returns true if the ray hit something
bool traceScreenSpaceRay(
    // Camera-space ray origin, which must be within the view volume
    float3 csOrig,
    // Unit length camera-space ray direction
    float3 csDir,
    // Number between 0 and 1 for how far to bump the ray in stride units
    // to conceal banding artifacts. Not needed if stride == 1.
    float jitter,
    // Pixel coordinates of the first intersection with the scene
    out float2 hitPixel,
    // Camera space location of the ray hit
    out float3 hitPoint)
{
    // Clip to the near plane
    float rayLength = ((csOrig.z + csDir.z * fFarClip) < 0.01) ?
    (0.01 - csOrig.z) / csDir.z : fFarClip;
    float3 csEndPoint = csOrig + csDir * rayLength;

    // Project into homogeneous clip space
    float4 H0 = mul(float4(csOrig, 1.0f), Projection);
    //H0.xy *= float2(fScreenWidth, fScreenHeight);
    float4 H1 = mul(float4(csEndPoint, 1.0f), Projection);
    //H1.xy *= float2(fScreenWidth, fScreenHeight);
    float k0 = 1.0f / H0.w;
    float k1 = 1.0f / H1.w;

    // The interpolated homogeneous version of the camera-space points
    float3 Q0 = csOrig * k0;
    float3 Q1 = csEndPoint * k1;

    // Screen-space endpoints
    float2 P0 = H0.xy * k0;
    float2 P1 = H1.xy * k1;

    // If the line is degenerate, make it cover at least one pixel
    // to avoid handling zero-pixel extent as a special case later
    P1 += (distanceSquared(P0, P1) < 0.0001f) ? float2(0.01f, 0.01f) : 0.0f;
    float2 delta = P1 - P0;

    // Permute so that the primary iteration is in x to collapse
    // all quadrant-specific DDA cases later
    bool permute = false;
    if (abs(delta.x) < abs(delta.y))
    {
        // This is a more-vertical line
        permute = true;
        delta = delta.yx;
        P0 = P0.yx;
        P1 = P1.yx;
    }

    float stepDir = sign(delta.x);
    float invdx = stepDir / delta.x;

    // Track the derivatives of Q and k
    float3 dQ = (Q1 - Q0) * invdx;
    float dk = (k1 - k0) * invdx;
    float2 dP = float2(stepDir, delta.y * invdx);

    // Scale derivatives by the desired pixel stride and then
    // offset the starting values by the jitter fraction
    float strideScale = 1.0f - min(1.0f, csOrig.z * 40);
    float stride = 1.0f + strideScale * 2;
    dP *= stride;
    dQ *= stride;
    dk *= stride;

    P0 += dP * jitter;
    Q0 += dQ * jitter;
    k0 += dk * jitter;

    // Slide P from P0 to P1, (now-homogeneous) Q from Q0 to Q1, k from k0 to k1
    float4 PQk = float4(P0, Q0.z, k0);
    float4 dPQk = float4(dP, dQ.z, dk);
    float3 Q = Q0;

    // Adjust end condition for iteration direction
    float end = P1.x * stepDir;

    float stepCount = 0.0f;
    float prevZMaxEstimate = csOrig.z;
    float rayZMin = prevZMaxEstimate;
    float rayZMax = prevZMaxEstimate;
    float sceneZMax = rayZMax + 100.0f;
    for (;
        ((PQk.x * stepDir) <= end) && (stepCount < 32) &&
        !intersectsDepthBuffer(sceneZMax, rayZMin, rayZMax) &&
        (sceneZMax != 0.0f);
        ++stepCount)
    {
        rayZMin = prevZMaxEstimate;
        rayZMax = (dPQk.z * 0.5f + PQk.z) / (dPQk.w * 0.5f + PQk.w);
        prevZMaxEstimate = rayZMax;
        if (rayZMin > rayZMax)
        {
            swap(rayZMin, rayZMax);
        }

        hitPixel = permute ? PQk.yx : PQk.xy;
        // You may need hitPixel.y = depthBufferSize.y - hitPixel.y; here if your vertical axis
        // is different than ours in screen space
        sceneZMax = linearDepthTexelFetch(hitPixel.xy);

        PQk += dPQk;
    }

    // Advance Q based on the number of steps
    Q.xy += dQ.xy * stepCount;
    hitPoint = Q * (1.0f / PQk.w);
    return intersectsDepthBuffer(sceneZMax, rayZMin, rayZMax);
}
bool TraceScreenSpaceRay(float3 startPoint, float3 rayDir, const float maxSteps, float maxDistance, out float2 hitPixel)
{
    float rayLength = ((startPoint.z + rayDir.z * maxDistance) > 0.01f) ? (0.01f - startPoint.z) / rayDir.z : maxDistance;
    float3 endPoint = startPoint + rayDir * rayLength;
     // Project into homogeneous clip space
    float4 H0 = mul(float4(startPoint, 1.0), Projection);
    H0.xy *= float2(fScreenWidth, fScreenHeight);
    float4 H1 = mul(float4(endPoint, 1.0), Projection);
    H1.xy *= float2(fScreenWidth, fScreenHeight);
    float k0 = 1.0 / H0.w, k1 = 1.0 / H1.w;
 
    // The interpolated homogeneous version of the camera-space points  
    float3 Q0 = startPoint * k0, Q1 = endPoint * k1;

    // Screen-space endpoints
    float2  P0 = H0.xy * k0,
            P1 = H1.xy * k1;
 
    // If the line is degenerate, make it cover at least one pixel
    // to avoid handling zero-pixel extent as a special case later
    P1 += float2(((distanceSquared(P0, P1) < 0.0001) ? 0.01 : 0.0).xx);
    float2 delta = P1 - P0;
 
    // Permute so that the primary iteration is in x to collapse
    // all quadrant-specific DDA cases later
    bool permute = false;
    if (abs(delta.x) < abs(delta.y))
    {
        // This is a more-vertical line
        permute = true;
        delta = delta.yx;
        P0 = P0.yx;
        P1 = P1.yx;
    }
 
    float stepDir = sign(delta.x);
    float invdx = stepDir / delta.x;
 
    // Track the derivatives of Q and k
    float3 dQ = (Q1 - Q0) * invdx;
    float dk = (k1 - k0) * invdx;
    float2 dP = float2(stepDir, delta.y * invdx);
 
    // Scale derivatives by the desired pixel stride and then
    // offset the starting values by the jitter fraction
/*    dP *= stride;
    dQ *= stride;
    dk *= stride;
    P0 += dP * jitter;
    Q0 += dQ * jitter;
    k0 += dk * jitter;
 */
    // Slide P from P0 to P1, (now-homogeneous) Q from Q0 to Q1, k from k0 to k1
    float3 Q = Q0;
 
    // Adjust end condition for iteration direction
    float end = P1.x * stepDir;
 
    float k = k0, stepCount = 0.0, prevZMaxEstimate = startPoint.z;
    float rayZMin = prevZMaxEstimate, rayZMax = prevZMaxEstimate;
    float sceneZMax = rayZMax + 100;
    for (float2 P = P0;
         ((P.x * stepDir) <= end) && (stepCount < maxSteps) &&
         ((rayZMax < sceneZMax - 0.001) || (rayZMin > sceneZMax)) &&
          (sceneZMax != 0);
         P += dP, Q.z += dQ.z, k += dk, ++stepCount)
    {
         
        rayZMin = prevZMaxEstimate;
        rayZMax = (dQ.z * 0.5 + Q.z) / (dk * 0.5 + k);
        prevZMaxEstimate = rayZMax;
        if (rayZMin > rayZMax)
        {
            float t = rayZMin;
            rayZMin = rayZMax;
            rayZMax = t;
        }
 
        hitPixel = permute ? P.yx : P;
        // You may need hitPixel.y = csZBufferSize.y - hitPixel.y; here if your vertical axis
        // is different than ours in screen space
        float4 NormalSpec = txGB1.Sample(samLinear, float2(hitPixel.x / fScreenWidth,1- hitPixel.y / fScreenHeight));
        sceneZMax = DecodeFloatRG(NormalSpec.zw);
    }
     
    // Advance Q based on the number of steps
    Q.xy += dQ.xy * stepCount;
    //hitPoint = Q * (1.0 / k);
    return (rayZMax >= sceneZMax - 0.001) && (rayZMin < sceneZMax);
}


float3 SSR(float3 texelPosition, float3 reflectDir,float roughness, out float fallback)
{
    float3 currentRay = 0;

    float3 nuv = 0;
    float L = 0.01;
    float fStep = SSRStep;
	float fStepScaling = 1.0f;
	fallback=0.0;
    float3 csRayStart = GetCameraSpacePos(texelPosition);
    float3 csRayDir = reflectDir;
   // return csRayStart;
    float2 uv;
    float3 hp;
    if (traceScreenSpaceRay(csRayStart, csRayDir, 0.5, uv, hp))
    {
        fallback = 0.0;
        return txGB0.Sample(samLinear, float2(uv.x / fScreenWidth, uv.y / fScreenHeight)).rgb;
    }
    else
    {
        fallback = 1.0;
        return float3(0, 0, 0);
    }
    //int maxIterations = min(SSRMaxIterations, 128);
    /*for (int i = 0; i < 32; i++)
    {
        currentRay = texelPosition + reflectDir * L;

        nuv = GetUV(currentRay);
        if (nuv.x < 0 || nuv.x > 1 || nuv.y < 0 || nuv.y > 1){
			fallback=1.0;
            return float3(0, 0, 0);
		}
        float4 NormalSpec = txGB1.Sample(samLinear, nuv.xy);
        float ViewZ = DecodeFloatRG(NormalSpec.zw);
        float3 newPosition = posFromDepth(ViewZ, nuv.xy).xyz;
        if (length(texelPosition - newPosition) <= SSRStep*1.1f)
        {
			fallback=1.0;
            return float3(0, 0, 0);
		}
        if (length(currentRay - newPosition) >= fStep - SSRStep)
		{
			L+=fStep;
			fStep*=fStepScaling;
		}
		else
		{
			break;
		}
    }
    float error0 = saturate(max(L - 0.011,0) / 0.044);
    float maxOutScreenRayDist = 0.03;
    float error1 = saturate(max(nuv.x - maxOutScreenRayDist, 0) / maxOutScreenRayDist * 0.25) *
                   saturate(max(nuv.y - maxOutScreenRayDist, 0) / maxOutScreenRayDist * 0.25) *
                   saturate(max(1 - nuv.x + maxOutScreenRayDist, 0) / maxOutScreenRayDist * 0.25) *
                   saturate(max(1 - nuv.y + maxOutScreenRayDist, 0) / maxOutScreenRayDist * 0.25);
	fallback = 1 - error0 * error1;*/
    //float3 refsample = txGB0.Sample(samLinear, nuv.xy).rgb;
    //return refsample; //* error0 * error1;
}

//--------------------------------------------------------------------------------------
// Pixel Shader
//--------------------------------------------------------------------------------------
float4 ReflectionPassPS(PSInput_Quad input) : SV_Target
{
    float4 NormalSpec = txGB1.Sample(samLinear, input.texCoordOut.xy);
    float ViewZ = DecodeFloatRG(NormalSpec.zw);
    ViewZ = ViewZ <= 0 ? fFarClip : ViewZ;
    float3 cameraSpacePos = viewPosFromDepth(ViewZ, input.texCoordOut.xy).xyz;
    float3 worldSpacePos = posFromDepth(ViewZ, input.texCoordOut.xy).xyz;
    float3 Normals = DecodeNormals(NormalSpec.xy);
    float2 Parameters = txGB2.Sample(samLinear, input.texCoordOut.xy).xy;
    float Roughness = 1 - Parameters.y;

    const float3 ViewPos = ViewInv[3].xyz;

    float3 ViewDir = normalize(worldSpacePos - ViewPos);
    float3 ReflDir = normalize(reflect(ViewDir, Normals));
    
    // simply return SSR for now, maybe will use DXR someday, but simple cubemap is way more possible 
    float Fallback;
    float FresnelCoeff = MicrofacetFresnel(vSunLightDir.xyz, normalize(vSunLightDir.xyz - ViewDir), Roughness);
    return  float4(/*SSR(worldSpacePos, ReflDir, Roughness, Fallback) * (1 - Fallback) * FresnelCoeff*/0,0,0, 1);
}
